{% extends "base.html" %}

{% block content %}
{% if page.nb_url %}
    <a href="{{ page.nb_url }}" title="Download Notebook" class="md-content__button md-icon">
        {% include ".icons/material/download.svg" %}
    </a>
{% endif %}
{% if page.file.src_uri %}
    {% if page.file.src_uri.endswith('.ipynb') %}
        <a href="https://colab.research.google.com/github/AutoResearch/autora/blob/main/docs/{{ page.file.src_uri }}"
        title="Open Notebook in Google Colab" class="md-content__button md-icon">
            {% include ".icons/material/play-circle-outline.svg" %}
        </a>
    {% endif %}
{% endif %}

{{ super() }}
{% endblock content %}
